@@ -32,58 +32,51 @@
 #include "e2k-uri.h"
 #include "e2k-utils.h"
 
-#include <e-util/e-xml-hash-utils.h>
+#include <libedataserver/e-xml-hash-utils.h>
 
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 
 struct _ExchangeHierarchyForeignPrivate {
-	gboolean checked_hide_private, checking_hide_private;
+	GMutex *hide_private_lock;
+	gboolean checked_hide_private;
 };
 
-typedef struct {
-	ExchangeAccountFolderCallback callback;
-	gpointer user_data;
-	gpointer op_data;
-} ExchangeHierarchyForeignDeferredOp;
-
 extern const char *exchange_localfreebusy_path;
 
-#define PARENT_TYPE EXCHANGE_TYPE_HIERARCHY_WEBDAV
-static ExchangeHierarchyWebDAVClass *parent_class = NULL;
+#define PARENT_TYPE EXCHANGE_TYPE_HIERARCHY_SOMEDAV
+static ExchangeHierarchySomeDAVClass *parent_class = NULL;
 
-static void async_create_folder (ExchangeHierarchy *hier, EFolder *parent,
-				 const char *name, const char *type,
-				 ExchangeAccountFolderCallback callback,
-				 gpointer user_data);
-static void async_remove_folder (ExchangeHierarchy *hier, EFolder *folder,
-				 ExchangeAccountFolderCallback callback,
-				 gpointer user_data);
-static void got_subtree  (E2kConnection *conn, SoupMessage *msg,
-			  E2kResult *results, int nresults,
-			  gpointer user_data);
-static void async_scan_subtree (ExchangeHierarchy *hier, EFolder *folder,
-				ExchangeAccountFolderCallback callback,
-				gpointer user_data);
-static void add_to_storage (ExchangeHierarchy *hier);
+static GPtrArray *get_hrefs (ExchangeHierarchySomeDAV *hsd);
+static ExchangeAccountFolderResult create_folder (ExchangeHierarchy *hier,
+						  EFolder *parent,
+						  const char *name,
+						  const char *type);
+static ExchangeAccountFolderResult remove_folder (ExchangeHierarchy *hier,
+						  EFolder *folder);
+static ExchangeAccountFolderResult scan_subtree (ExchangeHierarchy *hier,
+						 EFolder *folder);
 static void finalize (GObject *object);
 
 static void
 class_init (GObjectClass *object_class)
 {
-	ExchangeHierarchyClass *exchange_hierarchy_class =
+	ExchangeHierarchyClass *hierarchy_class =
 		EXCHANGE_HIERARCHY_CLASS (object_class);
+	ExchangeHierarchySomeDAVClass *somedav_class =
+		EXCHANGE_HIERARCHY_SOMEDAV_CLASS (object_class);
 
 	parent_class = g_type_class_ref (PARENT_TYPE);
 
 	/* virtual method override */
 	object_class->finalize = finalize;
 
-	exchange_hierarchy_class->add_to_storage      = add_to_storage;
-	exchange_hierarchy_class->async_create_folder = async_create_folder;
-	exchange_hierarchy_class->async_remove_folder = async_remove_folder;
-	exchange_hierarchy_class->async_scan_subtree  = async_scan_subtree;
+	hierarchy_class->create_folder  = create_folder;
+	hierarchy_class->remove_folder  = remove_folder;
+	hierarchy_class->scan_subtree   = scan_subtree;
+
+	somedav_class->get_hrefs        = get_hrefs;
 }
 
 static void
@@ -93,7 +86,7 @@ init (GObject *object)
 	ExchangeHierarchy *hier = EXCHANGE_HIERARCHY (object);
 
 	hfor->priv = g_new0 (ExchangeHierarchyForeignPrivate, 1);
-
+	hfor->priv->hide_private_lock = g_mutex_new ();
 	hier->hide_private_items = TRUE;
 }
 
@@ -102,6 +95,7 @@ finalize (GObject *object)
 {
 	ExchangeHierarchyForeign *hfor = EXCHANGE_HIERARCHY_FOREIGN (object);
 
+	g_mutex_free (hfor->priv->hide_private_lock);
 	g_free (hfor->priv);
 
 	G_OBJECT_CLASS (parent_class)->finalize (object);
@@ -109,47 +103,6 @@ finalize (GObject *object)
 
 E2K_MAKE_TYPE (exchange_hierarchy_foreign, ExchangeHierarchyForeign, class_init, init, PARENT_TYPE)
 
-static void
-got_delegation_props (E2kConnection *conn, SoupMessage *msg,
-		      E2kResult *results, int nresults,
-		      gpointer user_data)
-{
-	ExchangeHierarchyForeign *hfor = user_data;
-	ExchangeHierarchy *hier = user_data;
-	GPtrArray *entryids, *privflags;
-	GByteArray *entryid;
-	const char *my_dn, *delegate_dn;
-	int i;
-
-	hfor->priv->checking_hide_private = FALSE;
-	hfor->priv->checked_hide_private = TRUE;
-
-	if (!SOUP_ERROR_IS_SUCCESSFUL (msg->errorcode) || nresults != 1 ||
-	    !SOUP_ERROR_IS_SUCCESSFUL (results[0].status) || !results[0].props)
-		goto done;
-
-	my_dn = hier->account->legacy_exchange_dn;
-
-	entryids  = e2k_properties_get_prop (results[0].props, PR_DELEGATES_ENTRYIDS);
-	privflags = e2k_properties_get_prop (results[0].props, PR_DELEGATES_SEE_PRIVATE);
-	if (!entryids || !privflags)
-		goto done;
-
-	for (i = 0; i < entryids->len && i < privflags->len; i++) {
-		entryid = entryids->pdata[i];
-		delegate_dn = e2k_entryid_to_dn (entryid);
-
-		if (delegate_dn && !g_ascii_strcasecmp (delegate_dn, my_dn)) {
-			if (privflags->pdata[i] && atoi (privflags->pdata[i]))
-				hier->hide_private_items = FALSE;
-			break;
-		}
-	}
-
- done:
-	g_object_unref (hfor);
-}
-
 static const char *privacy_props[] = {
 	PR_DELEGATES_ENTRYIDS,
 	PR_DELEGATES_SEE_PRIVATE,
@@ -160,24 +113,65 @@ static void
 check_hide_private (ExchangeHierarchy *hier)
 {
 	ExchangeHierarchyForeign *hfor = EXCHANGE_HIERARCHY_FOREIGN (hier);
+	E2kContext *ctx;
+	E2kHTTPStatus status;
+	E2kResult *results;
+	int nresults, i;
+	GPtrArray *entryids, *privflags;
+	GByteArray *entryid;
+	const char *my_dn, *delegate_dn;
+	char *uri;
 
-	if (hfor->priv->checked_hide_private)
+	g_mutex_lock (hfor->priv->hide_private_lock);
+
+	if (hfor->priv->checked_hide_private) {
+		g_mutex_unlock (hfor->priv->hide_private_lock);
 		return;
-
-	if (!hfor->priv->checking_hide_private) {
-		hfor->priv->checking_hide_private = TRUE;
-
-		g_object_ref (hfor);
-		e_folder_exchange_bpropfind (hier->toplevel,
-					     &exchange_localfreebusy_path, 1, "0",
-					     privacy_props, n_privacy_props,
-					     got_delegation_props, hier);
 	}
 
-	while (hfor->priv->checking_hide_private)
-		g_main_context_iteration (NULL, TRUE);
+	uri = e2k_uri_concat (hier->account->home_uri,
+			      "NON_IPM_SUBTREE/Freebusy%20Data/LocalFreebusy.EML");
+	ctx = exchange_account_get_context (hier->account);
 
-	return;
+	status = e2k_context_propfind (ctx, NULL, uri,
+				       privacy_props, n_privacy_props,
+				       &results, &nresults);
+	g_free (uri);
+
+	hfor->priv->checked_hide_private = TRUE;
+
+	if (!E2K_HTTP_STATUS_IS_SUCCESSFUL (status) || nresults == 0) {
+		g_mutex_unlock (hfor->priv->hide_private_lock);
+		return;
+	}
+	if (!E2K_HTTP_STATUS_IS_SUCCESSFUL (results[0].status) ||
+	    !results[0].props || nresults > 1) {
+		e2k_results_free (results, nresults);
+		g_mutex_unlock (hfor->priv->hide_private_lock);
+		return;
+	}
+
+	entryids  = e2k_properties_get_prop (results[0].props,
+					     PR_DELEGATES_ENTRYIDS);
+	privflags = e2k_properties_get_prop (results[0].props,
+					     PR_DELEGATES_SEE_PRIVATE);
+	if (entryids && privflags) {
+		my_dn = hier->account->legacy_exchange_dn;
+		for (i = 0; i < entryids->len && i < privflags->len; i++) {
+			entryid = entryids->pdata[i];
+			delegate_dn = e2k_entryid_to_dn (entryid);
+
+			if (delegate_dn &&
+			    !g_ascii_strcasecmp (delegate_dn, my_dn) &&
+			    privflags->pdata[i] &&
+			    atoi (privflags->pdata[i]))
+				hier->hide_private_items = FALSE;
+			break;
+		}
+	}
+
+	e2k_results_free (results, nresults);
+	g_mutex_unlock (hfor->priv->hide_private_lock);
 }
 
 static void
@@ -201,91 +195,6 @@ hierarchy_foreign_cleanup (ExchangeHierarchy *hier)
 	exchange_hierarchy_removed_folder (hier, hier->toplevel);
 }
 
-static inline gboolean
-folder_is_unreadable (E2kProperties *props)
-{
-	char *access;
-
-	access = e2k_properties_get_prop (props, PR_ACCESS);
-	return !access || !atoi (access);
-}
-
-struct create_data {
-	ExchangeHierarchy *hier;
-	GList *uris;
-	ExchangeAccountFolderCallback callback;
-	gpointer user_data;
-};
-
-static void async_create_internal (E2kConnection *conn, struct create_data *cd);
-
-static void
-got_folder (E2kConnection *conn, SoupMessage *msg,
-	    E2kResult *results, int nresults,
-	    gpointer user_data)
-{
-	struct create_data *cd = user_data;
-	ExchangeAccountFolderResult result;
-	EFolder *folder = NULL;
-	int status;
-
-	if (SOUP_ERROR_IS_SUCCESSFUL (msg->errorcode)) {
-		if (nresults == 1)
-			status = results[0].status;
-		else
-			status = SOUP_ERROR_NOT_FOUND;
-	} else
-		status = msg->errorcode;
-
-	if (SOUP_ERROR_IS_SUCCESSFUL (status)) {
-		GPtrArray *folders = NULL;
-
-		exchange_hierarchy_webdav_parse_folders (
-			EXCHANGE_HIERARCHY_WEBDAV (cd->hier), msg,
-			cd->hier->toplevel, results, nresults, &folders);
-
-		if (folders && folders->len && !folder_is_unreadable (results[0].props)) {
-			folder = folders->pdata[0];
-			g_ptr_array_free (folders, TRUE);
-			exchange_hierarchy_new_folder (cd->hier, folder);
-			result = EXCHANGE_ACCOUNT_FOLDER_OK;
-		} else {
-			if (folders)
-				g_ptr_array_free (folders, TRUE);
-			result = EXCHANGE_ACCOUNT_FOLDER_PERMISSION_DENIED;
-		}
-	} else if (status == SOUP_ERROR_NOT_FOUND)
-		result = EXCHANGE_ACCOUNT_FOLDER_DOES_NOT_EXIST;
-	else if (status == SOUP_ERROR_UNAUTHORIZED)
-		result = EXCHANGE_ACCOUNT_FOLDER_PERMISSION_DENIED;
-	else {
-		g_warning ("hfor got_folder: %d %s", msg->errorcode, msg->errorphrase);
-		result = EXCHANGE_ACCOUNT_FOLDER_GENERIC_ERROR;
-	}
-
-	if (result != EXCHANGE_ACCOUNT_FOLDER_OK && cd->uris) {
-		async_create_internal (conn, cd);
-		return;
-	}
-
-	cd->callback (cd->hier->account, result, folder, cd->user_data);
-
-	/* If the hierarchy is now empty, then we must have just been
-	 * created but then the add failed. So remove it again.
-	 */
-	if (exchange_hierarchy_is_empty (cd->hier))
-		hierarchy_foreign_cleanup (cd->hier);
-
-	if (folder)
-		g_object_unref (folder);
-	g_object_unref (cd->hier);
-	while (cd->uris) {
-		g_free (cd->uris->data);
-		cd->uris = g_list_remove_link (cd->uris, cd->uris);
-	}
-	g_free (cd);
-}
-
 static const char *folder_props[] = {
 	E2K_PR_EXCHANGE_FOLDER_CLASS,
 	E2K_PR_HTTPMAIL_UNREAD_COUNT,
@@ -294,18 +203,42 @@ static const char *folder_props[] = {
 };
 static const int n_folder_props = sizeof (folder_props) / sizeof (folder_props[0]);
 
-static void
-async_create_internal (E2kConnection *conn, struct create_data *cd)
+static ExchangeAccountFolderResult
+find_folder (ExchangeHierarchy *hier, const char *uri, EFolder **folder_out)
 {
-	char *uri;
+	ExchangeHierarchyWebDAV *hwd = EXCHANGE_HIERARCHY_WEBDAV (hier);
+	E2kContext *ctx = exchange_account_get_context (hier->account);
+	E2kHTTPStatus status;
+	E2kResult *results;
+	int nresults;
+	EFolder *folder;
+	const char *access;
 
-	uri = cd->uris->data;
-	cd->uris = g_list_remove_link (cd->uris, cd->uris);
+	status = e2k_context_propfind (ctx, NULL, uri,
+				       folder_props, n_folder_props,
+				       &results, &nresults);
+	if (!E2K_HTTP_STATUS_IS_SUCCESSFUL (status)) 
+		return exchange_hierarchy_webdav_status_to_folder_result (status);
+	if (nresults == 0)
+		return EXCHANGE_ACCOUNT_FOLDER_DOES_NOT_EXIST;
 
-	e2k_connection_propfind (conn, uri, "0",
-				 folder_props, n_folder_props,
-				 got_folder, cd);
-	g_free (uri);
+	access = e2k_properties_get_prop (results[0].props, PR_ACCESS);
+	if (!access || !atoi (access))
+		return EXCHANGE_ACCOUNT_FOLDER_PERMISSION_DENIED;
+
+	folder = exchange_hierarchy_webdav_parse_folder (hwd, hier->toplevel,
+							 &results[0]);
+	e2k_results_free (results, nresults);
+
+	if (!folder)
+		return EXCHANGE_ACCOUNT_FOLDER_DOES_NOT_EXIST;
+
+	exchange_hierarchy_new_folder (hier, folder);
+	if (folder_out)
+		*folder_out = folder;
+	else
+		g_object_unref (folder);
+	return EXCHANGE_ACCOUNT_FOLDER_OK;
 }
 
 static struct {
@@ -323,235 +256,139 @@ static struct {
 	{ N_("Tasks"),		E2K_PR_STD_FOLDER_TASKS }
 };
 const static int n_std_folders = sizeof (std_folders) / sizeof (std_folders[0]);
-
-static void
-async_create_folder (ExchangeHierarchy *hier, EFolder *parent,
-		     const char *name, const char *type,
-		     ExchangeAccountFolderCallback callback,
-		     gpointer user_data)
+static ExchangeAccountFolderResult
+create_internal (ExchangeHierarchy *hier, EFolder *parent,
+		 const char *name, const char *type, EFolder **folder_out)
 {
-	struct create_data *cd;
+	ExchangeAccountFolderResult result;
 	char *literal_uri = NULL, *standard_uri = NULL;
 	const char *home_uri;
 	int i;
 
 	/* For now, no nesting */
-	if (parent != hier->toplevel || strchr (name + 1, '/')) {
-		callback (hier->account, EXCHANGE_ACCOUNT_FOLDER_GENERIC_ERROR,
-			  NULL, user_data);
-		return;
-	}
+	if (parent != hier->toplevel || strchr (name + 1, '/'))
+		return EXCHANGE_ACCOUNT_FOLDER_GENERIC_ERROR;
 
 	check_hide_private (hier);
 
 	home_uri = e_folder_exchange_get_internal_uri (hier->toplevel);
 	literal_uri = e2k_uri_concat (home_uri, name);
-	if (exchange_account_get_folder (hier->account, literal_uri))
-		goto already_exists;
+	if (exchange_account_get_folder (hier->account, literal_uri)) {
+		g_free (literal_uri);
+		if (exchange_hierarchy_is_empty (hier))
+			hierarchy_foreign_cleanup (hier);
+		return EXCHANGE_ACCOUNT_FOLDER_ALREADY_EXISTS;
+	}
 
 	for (i = 0; i < n_std_folders; i++) {
-		if (!g_ascii_strcasecmp (std_folders[i].name, name) ||
-		    !g_utf8_collate (_(std_folders[i].name), name))
-			break;
-	}
-	if (i < n_std_folders) {
+		if (g_ascii_strcasecmp (std_folders[i].name, name) != 0 &&
+		    g_utf8_collate (_(std_folders[i].name), name) != 0)
+			continue;
+
 		standard_uri = exchange_account_get_standard_uri_for (
 			hier->account, home_uri, std_folders[i].prop);
-		if (standard_uri) {
-			if (!strcmp (literal_uri, standard_uri)) {
-				g_free (standard_uri);
-				standard_uri = NULL;
-			} else if (exchange_account_get_folder (hier->account, standard_uri))
-				goto already_exists;
+		if (!standard_uri)
+			break;
+		if (!strcmp (literal_uri, standard_uri)) {
+			g_free (standard_uri);
+			standard_uri = NULL;
+			break;
 		}
+
+		if (exchange_account_get_folder (hier->account, standard_uri)) {
+			g_free (standard_uri);
+			g_free (literal_uri);
+			if (exchange_hierarchy_is_empty (hier))
+				hierarchy_foreign_cleanup (hier);
+			return EXCHANGE_ACCOUNT_FOLDER_ALREADY_EXISTS;
+		}
+
+		break;
 	}
 
-	cd = g_new0 (struct create_data, 1);
-	cd->hier = hier;
-	g_object_ref (hier);
-	cd->callback = callback;
-	cd->user_data = user_data;
+	result = find_folder (hier, literal_uri, folder_out);
+	if (result == EXCHANGE_ACCOUNT_FOLDER_DOES_NOT_EXIST && standard_uri)
+		result = find_folder (hier, standard_uri, folder_out);
 
-	cd->uris = g_list_prepend (NULL, literal_uri);
-	if (standard_uri)
-		cd->uris = g_list_prepend (cd->uris, standard_uri);
-
-	async_create_internal (exchange_account_get_connection (hier->account), cd);
-	return;
-
- already_exists:
 	g_free (literal_uri);
 	g_free (standard_uri);
+
+	/* If the hierarchy is now empty, then we must have just been
+	 * created but then the add failed. So remove it again.
+	 */
 	if (exchange_hierarchy_is_empty (hier))
 		hierarchy_foreign_cleanup (hier);
-	callback (hier->account, EXCHANGE_ACCOUNT_FOLDER_ALREADY_EXISTS,
-		  NULL, user_data);
+
+	return result;
 }
 
-static void
-async_remove_folder (ExchangeHierarchy *hier, EFolder *folder,
-		     ExchangeAccountFolderCallback callback,
-		     gpointer user_data)
+
+static ExchangeAccountFolderResult
+create_folder (ExchangeHierarchy *hier, EFolder *parent,
+	       const char *name, const char *type)
 {
-	if (folder == hier->toplevel)
-		hierarchy_foreign_cleanup (hier);
-	else
+	return create_internal (hier, parent, name, type, NULL);
+}
+
+static ExchangeAccountFolderResult
+remove_folder (ExchangeHierarchy *hier, EFolder *folder)
+{
+	if (folder != hier->toplevel)
 		exchange_hierarchy_removed_folder (hier, folder);
-	callback (hier->account, EXCHANGE_ACCOUNT_FOLDER_OK, NULL, user_data);
 
-	if (folder != hier->toplevel && exchange_hierarchy_is_empty (hier))
+	if (folder == hier->toplevel || exchange_hierarchy_is_empty (hier))
 		hierarchy_foreign_cleanup (hier);
+
+	return EXCHANGE_ACCOUNT_FOLDER_OK;
 }
 
-struct scan_data {
-	ExchangeHierarchy *hier;
-	ExchangeAccountFolderCallback callback;
-	gpointer user_data;
-};
-
-static void
-got_subtree (E2kConnection *conn, SoupMessage *msg,
-	     E2kResult *results, int nresults,
-	     gpointer user_data)
+static ExchangeAccountFolderResult
+scan_subtree (ExchangeHierarchy *hier, EFolder *folder)
 {
-	struct scan_data *sd = user_data;
-	ExchangeAccountFolderResult result;
-	GPtrArray *folders;
-	EFolder *folder;
-	E2kResult tmp;
-	int i;
+	ExchangeAccountFolderResult folder_result;
 
-	if (SOUP_ERROR_IS_SUCCESSFUL (msg->errorcode)) {
-		if (nresults == 0) {
-			result = EXCHANGE_ACCOUNT_FOLDER_DOES_NOT_EXIST;
-			goto done;
-		}
+	check_hide_private (hier);
 
-		/* If you have "folder visible" permission but nothing
-		 * else, you'll be able to fetch properties, but not
-		 * see anything in the folder. In that case, PR_ACCESS
-		 * will be 0, and we remove the folder from the list.
-		 */
-		for (i = 0; i < nresults; i++) {
-			if (!SOUP_ERROR_IS_SUCCESSFUL (results[i].status) ||
-			    folder_is_unreadable (results[i].props)) {
-				tmp = results[i];
-				results[i] = results[nresults - 1];
-				results[nresults - 1] = tmp;
-				i--;
-				nresults--;
-			}
-		}
+	folder_result = EXCHANGE_HIERARCHY_CLASS (parent_class)->scan_subtree (hier, folder);
 
-		if (nresults == 0) {
-			result = EXCHANGE_ACCOUNT_FOLDER_PERMISSION_DENIED;
-			goto done;
-		}
-	}
+	if (exchange_hierarchy_is_empty (hier))
+		hierarchy_foreign_cleanup (hier);
 
-	result = exchange_hierarchy_webdav_parse_folders (
-		EXCHANGE_HIERARCHY_WEBDAV (sd->hier), msg,
-		sd->hier->toplevel, results, nresults, &folders);
-	if (result == EXCHANGE_ACCOUNT_FOLDER_OK) {
-		for (i = 0; i < folders->len; i++) {
-			folder = folders->pdata[i];
-			exchange_hierarchy_new_folder (sd->hier, folder);
-			g_object_unref (folder);
-		}
-		g_ptr_array_free (folders, TRUE);
-	}
-
- done:
-	if (sd->callback)
-		sd->callback (sd->hier->account, result, sd->hier->toplevel, sd->user_data);
-
-	g_object_unref (sd->hier);
-	g_free (sd);
+	return folder_result;
 }
 
 static void
 add_href (ExchangeHierarchy *hier, EFolder *folder, gpointer hrefs)
 {
-	g_ptr_array_add (hrefs, g_strdup (e_folder_exchange_get_internal_uri (folder)));
+	g_ptr_array_add (hrefs, (char *)e_folder_exchange_get_internal_uri (folder));
 }
 
-static void
-async_scan_subtree_real (ExchangeHierarchy *hier, EFolder *folder,
-			 ExchangeAccountFolderCallback callback,
-			 gpointer user_data)
+static GPtrArray *
+get_hrefs (ExchangeHierarchySomeDAV *hsd)
 {
-	struct scan_data *sd;
 	GPtrArray *hrefs;
-	int i;
-
-	if (folder != hier->toplevel)
-		goto empty;
-
-	check_hide_private (hier);
 
 	hrefs = g_ptr_array_new ();
-	exchange_hierarchy_webdav_offline_scan_subtree (hier, add_href, hrefs);
-	if (!hrefs->len) {
-		g_ptr_array_free (hrefs, TRUE);
-		goto empty;
-	}
-
-	sd = g_new (struct scan_data, 1);
-	sd->hier = hier;
-	g_object_ref (hier);
-	sd->callback = callback;
-	sd->user_data = user_data;
-
-	e_folder_exchange_bpropfind (hier->toplevel,
-				     (const char **)hrefs->pdata, hrefs->len, "0",
-				     folder_props, n_folder_props,
-				     got_subtree, sd);
-
-	for (i = 0; i < hrefs->len; i++)
-		g_free (hrefs->pdata[i]);
-	g_ptr_array_free (hrefs, TRUE);
-	return;
-
- empty:
-	if (callback)
-		callback (hier->account, EXCHANGE_ACCOUNT_FOLDER_OK, folder, user_data);
-}
-
-static void
-async_scan_subtree (ExchangeHierarchy *hier, EFolder *folder,
-		    ExchangeAccountFolderCallback callback,
-		    gpointer user_data)
-{
-	/* Since we call async_scan_subtree_real in add_to_storage
-	 * below, this call must be redundant.
-	 */
-	if (callback)
-		callback (hier->account, EXCHANGE_ACCOUNT_FOLDER_OK, folder, user_data);
-}
-
-static void
-add_to_storage (ExchangeHierarchy *hier)
-{
-	EXCHANGE_HIERARCHY_CLASS (parent_class)->add_to_storage (hier);
-
-	async_scan_subtree_real (hier, hier->toplevel, NULL, NULL);
+	exchange_hierarchy_webdav_offline_scan_subtree (EXCHANGE_HIERARCHY (hsd), add_href, hrefs);
+	return hrefs;
 }
 
 /**
  * exchange_hierarchy_foreign_add_folder:
  * @hier: the hierarchy
  * @folder_name: the name of the folder to add
+ * @folder: on successful return, the created folder
  *
  * Adds a new folder to @hier.
+ *
+ * Return value: the folder result.
  **/
-void
-exchange_hierarchy_foreign_async_add_folder (ExchangeHierarchy *hier,
-					     const char *folder_name,
-					     ExchangeAccountFolderCallback callback,
-					     gpointer user_data)
+ExchangeAccountFolderResult
+exchange_hierarchy_foreign_add_folder (ExchangeHierarchy *hier,
+				       const char *folder_name,
+				       EFolder **folder)
 {
-	async_create_folder (hier, hier->toplevel, folder_name, NULL,
-			     callback, user_data);
+	return create_internal (hier, hier->toplevel, folder_name, NULL, folder);
 }
 
 static ExchangeHierarchy *
@@ -583,7 +420,7 @@ hierarchy_foreign_new (ExchangeAccount *account,
 					     internal_uri_prefix,
 					     owner_name, owner_email,
 					     source_uri,
-					     FALSE, "folder", 2);
+					     FALSE);
 
 	return EXCHANGE_HIERARCHY (hfor);
 }
